/*
 Copyright (C) 2007 Christian Dywan <christian@twotoasts.de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "main.h"

// -- search engines related

static GList* searchEngines_load(const gchar* filename, GError* *error)
{
    GList* searchEnginesNew = NULL;

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, filename, G_KEY_FILE_KEEP_COMMENTS, error);
    /*g_key_file_load_from_data_dirs(keyFile, sFilename, NULL
     , G_KEY_FILE_KEEP_COMMENTS, error);*/
    gchar** engines = g_key_file_get_groups(keyFile, NULL);
    guint i;
    for(i = 0; engines[i] != NULL; i++)
    {
        CSearchEngine* searchEngine = g_new0(CSearchEngine, 1);
        searchEngine->shortName = g_strdup(engines[i]);
        searchEngine->description = sokoke_key_file_get_string_default(keyFile
         , engines[i], "description", "", NULL);
        searchEngine->url = sokoke_key_file_get_string_default(keyFile
         , engines[i], "url", "", NULL);
        searchEngine->inputEncoding = sokoke_key_file_get_string_default(keyFile
         , engines[i], "input-encoding", "UTF-8", NULL);
        searchEngine->icon = sokoke_key_file_get_string_default(keyFile
         , engines[i], "icon", GTK_STOCK_FIND, NULL);
        searchEngine->keyword = sokoke_key_file_get_string_default(keyFile
         , engines[i], "keyword", "", NULL);
        searchEnginesNew = g_list_append(searchEnginesNew, searchEngine);
    }
    g_strfreev(engines);
    g_key_file_free(keyFile);
    return searchEnginesNew;
}

gboolean searchEngines_save(GList* searchEngines, const gchar* filename
 , GError** error)
{
    // Note: searchEngines is not freed here
    GKeyFile* keyFile = g_key_file_new();
    guint n = g_list_length(searchEngines);
    guint i; for(i = 0; i < n; i++)
    {
        CSearchEngine* searchEngine = (CSearchEngine*)g_list_nth_data(searchEngines, i);
        gchar* name = searchEngine->shortName; // FIXME: Can this be NULL?
        g_key_file_set_string(keyFile, name, "description", searchEngine->description);
        g_key_file_set_string(keyFile, name, "url", searchEngine->url);
        g_key_file_set_string(keyFile, name, "input-encoding", searchEngine->inputEncoding);
        g_key_file_set_string(keyFile, name, "icon", searchEngine->icon ? searchEngine->icon : "");
        g_key_file_set_string(keyFile, name, "keyword", searchEngine->keyword);
        g_free(searchEngine->shortName);
        g_free(searchEngine->url);
        g_free(searchEngine->inputEncoding);
        g_free(searchEngine->icon);
    }
    gboolean bSaved = sokoke_key_file_save_to_file(keyFile, filename, error);
    g_key_file_free(keyFile);

    return bSaved;
}

// -- configuration related

static CConfig* config_load(const gchar* filename, GError** error)
{
    CConfig* config = g_new0(CConfig, 1);

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, filename, G_KEY_FILE_KEEP_COMMENTS, error);
    /*g_key_file_load_from_data_dirs(keyFile, sFilename, NULL
     , G_KEY_FILE_KEEP_COMMENTS, error);*/
    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "browser", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "browser", key, default, NULL)
    GET_INT(config->startup, "Startup", CONFIG_STARTUP_HOMEPAGE);
    GET_STR(config->homepage, "Homepage", "http://www.google.com");
    GET_STR(config->locationSearch, "LocationSearch", "http://www.google.com/search?q=%s");
    GET_INT(config->toolbarNavigation, "ToolbarNavigation", TRUE);
    GET_INT(config->toolbarBookmarks, "ToolbarBookmarks", FALSE);
    GET_INT(config->toolbarStatus, "ToolbarStatus", TRUE);
    //GET_INT(config->toolbarDownloads, "ToolbarDownloads", FALSE);
    GET_INT(config->toolbarStyle, "ToolbarStyle", CONFIG_TOOLBAR_DEFAULT);
    GET_INT(config->toolbarSmall, "ToolbarSmall", FALSE);
    GET_INT(config->toolbarWebSearch, "ToolbarWebSearch", TRUE);
    GET_INT(config->toolbarNewTab, "ToolbarNewTab", TRUE);
    GET_INT(config->toolbarClosedTabs, "ToolbarClosedTabs", TRUE);
    GET_INT(config->panelShow, "PanelShow", FALSE);
    GET_INT(config->panelActive, "PanelActive", 0);
    GET_STR(config->panelPageholder, "PanelPageholder", "http://www.google.com");
    GET_INT(config->tabSize, "TabSize", 10);
    GET_INT(config->tabClose, "TabClose", TRUE);
    GET_INT(config->newPages, "NewPages", CONFIG_NEWPAGES_TAB_NEW);
    GET_INT(config->openTabsInTheBackground, "OpenTabsInTheBackground", FALSE);
    GET_INT(config->openPopupsInTabs, "OpenPopupsInTabs", FALSE);
    #undef GET_INT
    #undef GET_STR

    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "content", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "content", key, default, NULL)
    GET_INT(config->loadImagesAutomatically, "LoadImagesAutomatically", TRUE);
    GET_INT(config->shrinkImagesToFit, "ShrinkImagesTooFit", TRUE);
    GET_INT(config->resizableTextAreas, "ResizableTextAreas", FALSE);
    GET_INT(config->enableJavaScript, "EnableJavaScript", TRUE);
    GET_INT(config->enableJava, "EnableJava", TRUE);
    GET_INT(config->enablePlugins, "EnablePlugins", TRUE);
    #undef GET_INT
    #undef GET_STR

    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "session", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "session", key, default, NULL)
    GET_INT(config->rememberWinMetrics  , "RememberWinMetrics", TRUE);
    GET_INT(config->winLeft  , "WinLeft", 0);
    GET_INT(config->winTop   , "WinTop", 0);
    GET_INT(config->winWidth , "WinWidth", 0);
    GET_INT(config->winHeight, "WinHeight", 0);
    GET_INT(config->winPanelPos, "WinPanelPos", 0);
    GET_INT(config->searchEngine, "SearchEngine", 0);
    #undef GET_INT
    #undef GET_STR

    config->protocols_names = g_ptr_array_new();
    g_datalist_init(&config->protocols_commands);
    gchar** protocols;
    if((protocols = g_key_file_get_keys(keyFile, "protocols", NULL, NULL)))
    {
        guint i; for(i = 0; protocols[i] != NULL; i++)
        {
            gchar* sCommand = g_key_file_get_string(keyFile, "protocols"
             , protocols[i], NULL);
            g_ptr_array_add(config->protocols_names, (gpointer)protocols[i]);
            g_datalist_set_data_full(&config->protocols_commands
             , protocols[i], sCommand, g_free);
        }
        g_free(protocols);
    }

    g_key_file_free(keyFile);
    return config;
}

gboolean config_save(CConfig* config, const gchar* filename, GError** error)
{
    GKeyFile* keyFile = g_key_file_new();

    g_key_file_set_integer(keyFile, "browser", "Startup", config->startup);
    g_key_file_set_string (keyFile, "browser", "Homepage", config->homepage);
    g_key_file_set_string (keyFile, "browser", "LocationSearch", config->locationSearch);
    g_key_file_set_integer(keyFile, "browser", "ToolbarNavigation", config->toolbarNavigation);
    g_key_file_set_integer(keyFile, "browser", "ToolbarBookmarks", config->toolbarBookmarks);
    //g_key_file_set_integer(keyFile, "browser", "ToolbarDownloads", config->toolbarDownloads);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStatus", config->toolbarStatus);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStyle", config->toolbarStyle);
    g_key_file_set_integer(keyFile, "browser", "ToolbarSmall", config->toolbarSmall);
    g_key_file_set_integer(keyFile, "browser", "ToolbarWebSearch", config->toolbarWebSearch);
    g_key_file_set_integer(keyFile, "browser", "ToolbarNewTab", config->toolbarNewTab);
    g_key_file_set_integer(keyFile, "browser", "ToolbarClosedTabs", config->toolbarClosedTabs);
    g_key_file_set_integer(keyFile, "browser", "PanelShow", config->panelShow);
    g_key_file_set_integer(keyFile, "browser", "PanelActive", config->panelActive);
    g_key_file_set_string (keyFile, "browser", "PanelPageholder", config->panelPageholder);
    g_key_file_set_integer(keyFile, "browser", "TabSize", config->tabSize);
    g_key_file_set_integer(keyFile, "browser", "TabClose", config->tabClose);
    g_key_file_set_integer(keyFile, "browser", "NewPages", config->newPages);
    g_key_file_set_integer(keyFile, "browser", "OpenTabsInTheBackground", config->openTabsInTheBackground);
    g_key_file_set_integer(keyFile, "browser", "OpenPopupsInTabs", config->openPopupsInTabs);

    g_key_file_set_integer(keyFile, "content", "LoadImagesAutomatically", config->loadImagesAutomatically);
    g_key_file_set_integer(keyFile, "content", "ShrinkImagesToFit", config->shrinkImagesToFit);
    g_key_file_set_integer(keyFile, "content", "ResizableTextAreas", config->resizableTextAreas);
    g_key_file_set_integer(keyFile, "content", "EnableJavaScript", config->enableJavaScript);
    g_key_file_set_integer(keyFile, "content", "EnableJava", config->enableJava);
    g_key_file_set_integer(keyFile, "content", "EnablePlugins", config->enablePlugins);

    g_key_file_set_integer(keyFile, "session", "RememberWinMetrics", config->rememberWinMetrics);
    g_key_file_set_integer(keyFile, "session", "WinLeft", config->winLeft);
    g_key_file_set_integer(keyFile, "session", "WinTop", config->winTop);
    g_key_file_set_integer(keyFile, "session", "WinWidth", config->winWidth);
    g_key_file_set_integer(keyFile, "session", "WinHeight", config->winHeight);
    g_key_file_set_integer(keyFile, "session", "WinPanelPos", config->winPanelPos);
    g_key_file_set_integer(keyFile, "session", "SearchEngine", config->searchEngine);

    guint i; for(i = 0; i < config->protocols_names->len; i++)
    {
        gchar* protocol = (gchar*)g_ptr_array_index(config->protocols_names, i);
        gchar* command = g_datalist_get_data(&config->protocols_commands, protocol);
        g_key_file_set_string(keyFile, "protocols", protocol, command);
        g_free(protocol);
    }
    g_datalist_clear(&config->protocols_commands);
    g_ptr_array_free(config->protocols_names, TRUE);

    gboolean saved = sokoke_key_file_save_to_file(keyFile, filename, error);
    g_key_file_free(keyFile);

    return saved;
}

// -- stock icons

void stock_items_init(void)
{
    static GtkStockItem items[] =
    {
        { STOCK_LOCK_OPEN },
        { STOCK_LOCK_SECURE },
        { STOCK_LOCK_BROKEN },
        { STOCK_SCRIPT },
        { STOCK_THEME },
        { STOCK_USER_TRASH },

        { STOCK_BOOKMARK,       "Bookmark", 0, 0, NULL },
        { STOCK_BOOKMARK_NEW,   "New bookmark", 0, 0, NULL },
        { STOCK_BOOKMARKS,      "_Bookmarks", 0, 0, NULL },
        { STOCK_DOWNLOADS,      "_Downloads", 0, 0, NULL },
        { STOCK_CONSOLE,        "_Console", 0, 0, NULL },
        { STOCK_EXTENSIONS,     "_Extensions", 0, 0, NULL },
        { STOCK_FORM_FILL,      "_Form fill", 0, 0, NULL },
        { STOCK_HISTORY,        "History", 0, 0, NULL },
        { STOCK_HOMEPAGE,       "Homepage", 0, 0, NULL },
        { STOCK_LOCATION,       "Location entry", 0, 0, NULL },
        { STOCK_NEWSFEED,       "Newsfeed", 0, 0, NULL },
        { STOCK_PLUGINS,        "_Plugins", 0, 0, NULL },
        { STOCK_POPUPS_BLOCKED, "Blocked popups", 0, 0, NULL },
        { STOCK_SOURCE_VIEW,    "View source", 0, 0, NULL },
        { STOCK_TAB_CLOSE,      "Close tab", 0, 0, NULL },
        { STOCK_TAB_NEW,        "New tab", 0, 0, NULL },
        { STOCK_WINDOW_CLOSE,   "Close window", 0, 0, NULL },
        { STOCK_WINDOW_NEW,     "New window", 0, 0, NULL },
        #if !GTK_CHECK_VERSION(2, 10, 0)
        { GTK_STOCK_SELECT_ALL, "Select _All", 0, 0, (gchar*)"gtk20" },
        #endif
    };
    GtkIconFactory* factory = gtk_icon_factory_new();
    guint i; for(i = 0; i < (guint)G_N_ELEMENTS(items); i++)
    {
        GtkIconSource* iconSource = gtk_icon_source_new();
        gtk_icon_source_set_icon_name(iconSource, items[i].stock_id);
        GtkIconSet* iconSet = gtk_icon_set_new();
        gtk_icon_set_add_source(iconSet, iconSource);
        gtk_icon_source_free(iconSource);
        gtk_icon_factory_add(factory, items[i].stock_id, iconSet);
        gtk_icon_set_unref(iconSet);
    }
    gtk_stock_add_static(items, G_N_ELEMENTS(items));
    gtk_icon_factory_add_default(factory);
    g_object_unref(factory);
}

// -- main function

int main(int argc, char** argv)
{
    g_set_application_name(PACKAGE_NAME);

    // Parse cli options
    gint repeats = 2;
    gboolean version = FALSE;
    GOptionEntry entries[] =
    {
     { "repeats", 'r', 0, G_OPTION_ARG_INT, &repeats, "An unused value", "N" },
     { "version", 'v', 0, G_OPTION_ARG_NONE, &version, "Display program version", NULL }
    };

    GError* error = NULL;
    if(!gtk_init_with_args(&argc, &argv, "[URI]", entries
     , NULL/*GETTEXT_PACKAGE*/, &error))
    {
        g_error_free(error);
        return 1;
    }

    if(version)
    {
        g_print(PACKAGE_STRING " - Copyright (c) 2007 Christian Dywan\n\n"
         "GTK+2:      " GTK_VER "\n"
         "WebKit:     " WEBKIT_VER "\n"
         "Libsexy:    " LIBSEXY_VER "\n"
         "libXML2:    " LIBXML_VER "\n"
         "GetText:    N/A\n"
         "\n"
         "Debugging:  " SOKOKE_DEBUG_ "\n"
         "\n"
         "Please report comments, suggestions and bugs to:\n"
         "\t" PACKAGE_BUGREPORT "\n"
         "Check for new versions at:\n"
         "\thttp://software.twotoasts.de\n");
        return 0;
    }

    // Load configuration files
    GString* errorMessages = g_string_new(NULL);
    // TODO: What about default config in a global config folder?
    gchar* configPath = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(configPath, 0755);
    gchar* configFile = g_build_filename(configPath, "config", NULL);
    error = NULL;
    config = config_load(configFile, &error);
    g_free(configFile);
    if(error != NULL)
    {
        if(error->code != G_FILE_ERROR_NOENT)
            g_string_append_printf(errorMessages
             , "Configuration was not loaded. %s\n", error->message);
        g_error_free(error);
    }
    configFile = g_build_filename(configPath, "accels", NULL);
    gtk_accel_map_load(configFile);
    g_free(configFile);
    configFile = g_build_filename(configPath, "search", NULL);
    error = NULL;
    searchEngines = searchEngines_load(configFile, &error);
    g_free(configFile);
    if(error != NULL)
    {
        // FIXME: We may have a "file empty" error, how do we recognize that?
        /*if(error->code != G_FILE_ERROR_NOENT)
            g_string_append_printf(errorMessages
             , "Notice: No search engines loaded. %s\n", error->message);*/
        g_error_free(error);
    }
    configFile = g_build_filename(configPath, "bookmarks.xbel", NULL);
    bookmarks = xbel_folder_new();
    error = NULL;
    if(!xbel_folder_from_file(bookmarks, configFile, &error))
    {
        if(error->code != G_FILE_ERROR_NOENT)
            g_string_append_printf(errorMessages
             , "Bookmarks couldn't be loaded. %s\n", error->message);
        g_error_free(error);
    }
    g_free(configFile);
    XBELitem* _session = xbel_folder_new();
    if(config->startup == CONFIG_STARTUP_SESSION)
    {
        configFile = g_build_filename(configPath, "session.xbel", NULL);
        error = NULL;
        if(!xbel_folder_from_file(_session, configFile, &error))
        {
            if(error->code != G_FILE_ERROR_NOENT)
                g_string_append_printf(errorMessages
                 , "Session couldn't be loaded. %s\n", error->message);
            g_error_free(error);
        }
        g_free(configFile);
    }

    // In case of errors
    if(errorMessages->len)
    {
        GtkWidget* dialog = gtk_message_dialog_new(NULL
         , 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_NONE
         , "The following errors occured.");
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(dialog), FALSE);
        gtk_window_set_title(GTK_WINDOW(dialog), g_get_application_name());
        // FIXME: Use custom program icon
        gtk_window_set_icon_name(GTK_WINDOW(dialog), "web-browser");
        gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog)
         , "%s", errorMessages->str);
        gtk_dialog_add_buttons(GTK_DIALOG(dialog)
         , GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
         , "_Ignore", GTK_RESPONSE_ACCEPT
         , NULL);
        if(gtk_dialog_run(GTK_DIALOG(dialog)) != GTK_RESPONSE_ACCEPT)
        {
            // FIXME: free config
            // FIXME: search engines
            xbel_item_free(bookmarks);
            xbel_item_free(_session);
            g_string_free(errorMessages, TRUE);
            return 0;
        }
        gtk_widget_destroy(dialog);
        /* FIXME: Since we will overwrite files that could not be loaded
                  , would we want to make backups? */
    }
    g_string_free(errorMessages, TRUE);

    if(!xbel_folder_get_n_items(_session))
    {
        XBELitem* item = xbel_bookmark_new();
        if(config->startup == CONFIG_STARTUP_HOMEPAGE)
            xbel_bookmark_set_href(item, config->homepage);
        else
            xbel_bookmark_set_href(item, "about:blank");
        xbel_folder_append_item(_session, item);
    }
    g_free(configPath);

    accel_group = gtk_accel_group_new();
    stock_items_init();
    webkit_init();
    browsers = NULL;

    // TODO: Handle any number of separate uris from argv 
    // Open as many tabs as we have uris, seperated by pipes
    gchar* uri = argc > 1 ? strtok(g_strdup(argv[1]), "|") : NULL;
    while(uri != NULL)
    {
        XBELitem* item = xbel_bookmark_new();
        gchar* uriReady = magic_uri(uri, FALSE);
        xbel_bookmark_set_href(item, uriReady); g_free(uriReady);
        xbel_folder_append_item(_session, item);
        uri = strtok(NULL, "|");
    } g_free(uri);

    session = xbel_folder_new();
    CBrowser* browser = NULL;
    guint i; for(i = 0; i < xbel_folder_get_n_items(_session); i++)
    {
        XBELitem* item = xbel_folder_get_nth_item(_session, i);
        browser = browser_new(browser);
        webView_open(browser->webView, xbel_bookmark_get_href(item));
    }
    xbel_item_free(_session);

    gtk_main();

    // Save configuration files
    configPath = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(configPath, 0755);
    configFile = g_build_filename(configPath, "config", NULL);
    error = NULL;
    if(!config_save(config, configFile, &error))
    {
        g_warning("Configuration couldn't be saved. %s", error->message);
        g_error_free(error);
    }
    g_free(configFile);
    configFile = g_build_filename(configPath, "accels", NULL);
    gtk_accel_map_save(configFile);
    g_free(configFile);
    configFile = g_build_filename(configPath, "search", NULL);
    error = NULL;
    if(!searchEngines_save(searchEngines, configFile, &error))
    {
        g_warning("Search engines couldn't be saved. %s", error->message);
        g_error_free(error);
    }
    g_list_free(searchEngines);
    g_free(configFile);
    configFile = g_build_filename(configPath, "bookmarks.xbel", NULL);
    error = NULL;
    if(!xbel_folder_to_file(bookmarks, configFile, &error))
    {
        g_warning("Bookmarks couldn't be saved. %s", error->message);
        g_error_free(error);
    }
    xbel_item_free(bookmarks);
    g_free(configFile);
    if(config->startup == CONFIG_STARTUP_SESSION)
    {
        configFile = g_build_filename(configPath, "session.xbel", NULL);
        error = NULL;
        if(!xbel_folder_to_file(session, configFile, &error))
        {
            g_warning("Session couldn't be saved. %s", error->message);
            g_error_free(error);
        }
        g_free(configFile);
    }
    xbel_item_free(session);
    g_free(configPath);
    return 0;
}
