/*
 Copyright (C) 2007 Christian Dywan <christian@twotoasts.de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/**
 * This is an implementation of XBEL based on Glib and libXML2.
 *
 * Design Goals:
 *  - XBELitem is the only opaque public data structure.
 *  - The interface should be intuitive and familiar to Glib users.
 *  - There should be no public exposure of libXML2 specific code.
 *  - Bookmarks should actually be easily exchangeable between programs.
 *
 * TODO:
 *  - Support info > metadata, alias, added, modified, visited
 *  - Compatibility: The Nokia 770 *requires* metadata and folder
 *  - Compatibility: Kazehakase's bookmarks
 *  - Compatibility: Epiphany's bookmarks
 *  - XML Indentation
 *  - Optional xbel_import_file
 *  - Optional xbel_export_file
 *  - Consider a more liberate license
 **/

#include "xbel.h"

// Private: Create a new item of a certain type
static XBELitem* xbel_item_new(XBELitemType type)
{
    XBELitem* item = g_new(XBELitem, 1);
    item->parent = NULL;
    item->type = type;
    if(type == XBEL_ITEM_FOLDER)
    {
        item->items = NULL;
        item->folded = TRUE;
    }
    if(type != XBEL_ITEM_SEPARATOR)
    {
        item->title = NULL;
        item->desc  = NULL;
    }
    if(type == XBEL_ITEM_BOOKMARK) item->href = g_strdup("");
    return item;
}

/**
 * xbel_bookmark_new:
 * Create a new empty bookmark.
 *
 * Return value: a newly allocated bookmark
 **/
XBELitem* xbel_bookmark_new(void)
{
    return xbel_item_new(XBEL_ITEM_BOOKMARK);
}

static XBELitem* xbel_bookmark_from_xmlNodePtr(xmlNodePtr cur)
{
    g_return_val_if_fail(cur != NULL, NULL);
    XBELitem* bookmark = xbel_bookmark_new();
    xmlChar* key = xmlGetProp(cur, (xmlChar*)"href");
    xbel_bookmark_set_href(bookmark, (gchar*)key);
    cur = cur->xmlChildrenNode;
    while(cur != NULL)
    {
        if(!xmlStrcmp(cur->name, (const xmlChar*)"title"))
        {
         xmlChar* key = xmlNodeGetContent(cur);
         bookmark->title = (gchar*)g_strstrip((gchar*)key);
        }
        else if(!xmlStrcmp(cur->name, (const xmlChar*)"desc"))
        {
         xmlChar* key = xmlNodeGetContent(cur);
         bookmark->desc = (gchar*)g_strstrip((gchar*)key);
        }
        cur = cur->next;
    }
    return bookmark;
}

/**
 * xbel_separator_new:
 * Create a new separator.
 *
 * The returned item must be freed eventually.
 *
 * Return value: a newly allocated separator.
 **/
XBELitem* xbel_separator_new(void)
{
    return xbel_item_new(XBEL_ITEM_SEPARATOR);
}

/**
 * xbel_folder_new:
 * Create a new empty folder.
 *
 * The returned item must be freed eventually.
 *
 * Return value: a newly allocated folder.
 **/
XBELitem* xbel_folder_new(void)
{
    return xbel_item_new(XBEL_ITEM_FOLDER);
}

// Private: Create a folder from an xmlNodePtr
static XBELitem* xbel_folder_from_xmlNodePtr(xmlNodePtr cur)
{
    g_return_val_if_fail(cur != NULL, NULL);
    XBELitem* folder = xbel_folder_new();
    xmlChar* key = xmlGetProp(cur, (xmlChar*)"folded");
    if(key)
    {
        if(!g_ascii_strncasecmp((gchar*)key, "yes", 3))
            folder->folded = TRUE;
        else if(!g_ascii_strncasecmp((gchar*)key, "no", 2))
            folder->folded = FALSE;
        else
            g_warning("XBEL: Unknown value for folded.");
        xmlFree(key);
    }
    cur = cur->xmlChildrenNode;
    while(cur)
    {
        if(!xmlStrcmp(cur->name, (const xmlChar*)"title"))
        {
            xmlChar* key = xmlNodeGetContent(cur);
            folder->title = (gchar*)g_strstrip((gchar*)key);
        }
        else if(!xmlStrcmp(cur->name, (const xmlChar*)"desc"))
        {
            xmlChar* key = xmlNodeGetContent(cur);
            folder->desc = (gchar*)g_strstrip((gchar*)key);
        }
        else if(!xmlStrcmp(cur->name, (const xmlChar*)"folder"))
        {
            XBELitem* item = xbel_folder_from_xmlNodePtr(cur);
            item->parent = (struct XBELitem*)folder;
            folder->items = g_list_prepend(folder->items, item);
        }
     else if(!xmlStrcmp(cur->name, (const xmlChar*)"bookmark"))
     {
         XBELitem* item = xbel_bookmark_from_xmlNodePtr(cur);
         item->parent = (struct XBELitem*)folder;
         folder->items = g_list_prepend(folder->items, item);
     }
     else if(!xmlStrcmp(cur->name, (const xmlChar*)"separator"))
     {
         XBELitem* item = xbel_separator_new();
         item->parent = (struct XBELitem*)folder;
         folder->items = g_list_prepend(folder->items, item);
     }
        cur = cur->next;
    }
    // Prepending and reversing is faster than appending
    folder->items = g_list_reverse(folder->items);
    return folder;
}

// Private: Loads the contents from an xmlNodePtr into a folder.
static gboolean xbel_folder_from_xmlDocPtr(XBELitem* folder, xmlDocPtr doc)
{
    g_return_val_if_fail(doc != NULL, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(xbel_folder_get_n_items(folder))
        g_warning("XBEL: The folder is not empty.");
    // FIXME: If the folder is not empty, the resulting order is undefined
    xmlNodePtr cur = xmlDocGetRootElement(doc);
    xmlChar* version = xmlGetProp(cur, (xmlChar*)"version");
    if(xmlStrcmp(version, (xmlChar*)"1.0"))
        g_warning("XBEL version is not 1.0.");
    xmlFree(version);
    folder->title = (gchar*)xmlGetProp(cur, (xmlChar*)"title");
    folder->desc = (gchar*)xmlGetProp(cur, (xmlChar*)"desc");
    if((cur = xmlDocGetRootElement(doc)) == NULL)
    {
        // Empty document
        return FALSE;
    }
    if(xmlStrcmp(cur->name, (const xmlChar*)"xbel"))
    {
        // Wrong document type
        return FALSE;
    }
    cur = cur->xmlChildrenNode;
    while(cur != NULL)
    {
        XBELitem* item = NULL;
        if(!xmlStrcmp(cur->name, (const xmlChar*)"folder"))
         item = xbel_folder_from_xmlNodePtr(cur);
        else if(!xmlStrcmp(cur->name, (const xmlChar*)"bookmark"))
         item = xbel_bookmark_from_xmlNodePtr(cur);
        else if(!xmlStrcmp(cur->name, (const xmlChar*)"separator"))
         item = xbel_separator_new();
        /*else if(!xmlStrcmp(cur->name, (const xmlChar*)"info"))
         item = xbel_parse_info(xbel, cur);*/
        if(item != NULL)
        {
            item->parent = (struct XBELitem*)folder;
            folder->items = g_list_prepend(folder->items, item);
        }
        cur = cur->next;
    }
    // Prepending and reversing is faster than appending
    folder->items = g_list_reverse(folder->items);
    return TRUE;
}

/**
 * xbel_item_free:
 * @item: a valid item
 *
 * Free an XBELitem which is not contained in a folder.
 *
 * Freeing items contained in a folder will fail.
 **/
void xbel_item_free(XBELitem* item)
{
    g_return_if_fail(item);
    g_return_if_fail(!xbel_item_get_parent(item));
    if(xbel_item_is_folder(item))
    {
        guint n = xbel_folder_get_n_items(item);
        guint i; for(i = 0; i < n; i++)
        {
            XBELitem* _item = xbel_folder_get_nth_item(item, i);
            _item->parent = NULL;
            xbel_item_free(_item);
        }
        g_list_free(item->items);
    }
    if(item->type != XBEL_ITEM_SEPARATOR)
    {
        g_free(item->title);
        g_free(item->desc);
    }
    if(item->type == XBEL_ITEM_BOOKMARK)
        g_free(item->href);
    g_free(item);
}

/**
 * xbel_item_copy:
 * @item: the item to copy
 *
 * Copy an XBELitem.
 *
 * The returned item must be freed eventually.
 *
 * Return value: a copy of @item
 **/
XBELitem* xbel_item_copy(XBELitem* item)
{
    // FIXME: Should undergo a few tests
    XBELitem* copy = xbel_folder_new();
    if(xbel_item_is_folder(item))
    {
        guint n = xbel_folder_get_n_items(item);
        guint i; for(i = 0; i < n; i++)
        {
            XBELitem* _item = xbel_folder_get_nth_item(item, i);
            xbel_folder_append_item(copy, xbel_item_copy(_item));
        }
    }
    if(item->type != XBEL_ITEM_SEPARATOR)
    {
        xbel_item_set_title(copy, item->title);
        xbel_item_set_desc(copy, item->desc);
    }
    if(item->type == XBEL_ITEM_BOOKMARK)
        xbel_bookmark_set_href(copy, item->href);
    return copy;
}

/**
 * xbel_folder_append_item:
 * @folder: a folder
 * @item: the item to append
 *
 * Append the given item to a folder.
 *
 * The item is actually moved and must not be contained in another folder.
 *
 **/
void xbel_folder_append_item(XBELitem* folder, XBELitem* item)
{
    g_return_if_fail(!xbel_item_get_parent(item));
    g_return_if_fail(xbel_item_is_folder(folder));
    // FIXME: Should undergo a few tests
    item->parent = (struct XBELitem*)folder;
    folder->items = g_list_append(folder->items, item);
}

/**
 * xbel_folder_remove_item:
 * @folder: a folder
 * @item:   the item to remove
 *
 * Remove the given @item from a @folder.
 *
 * Return value: the @item on success or %NULL if the @item wasn't found
 **/
XBELitem* xbel_folder_remove_item(XBELitem* folder, XBELitem* item)
{
    g_return_val_if_fail(item, NULL);
    g_return_val_if_fail(xbel_item_get_parent(folder) != item, NULL);
    // FIXME: Should undergo a few tests
    item->parent = NULL;
    // Fortunately we know that items are unique
    folder->items = g_list_remove(folder->items, item);
    // FIXME: We don't actually know wether this fails
    return item;
}

/**
 * xbel_folder_get_n_items:
 * @folder: a folder
 *
 * Retrieve the number of items contained in the given @folder.
 *
 * Return value: number of items
 **/
guint xbel_folder_get_n_items(XBELitem* folder)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), 0);
    return g_list_length(folder->items);
}

/**
 * xbel_folder_get_nth_item:
 * @folder: a folder
 * @n: the index of an item
 *
 * Retrieve an item contained in the given @folder by its index.
 *
 * Return value: the item at the given index or %NULL
 **/
XBELitem* xbel_folder_get_nth_item(XBELitem* folder, guint n)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), NULL);
    return (XBELitem*)g_list_nth_data(folder->items, n);
}

/**
 * xbel_folder_get_folded:
 * @folder: A folder.
 *
 * ...
 *
 * Return value: The type of the given @folder.
 **/
gboolean xbel_folder_get_folded(XBELitem* folder)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), TRUE);
    return folder->folded;
}

/**
 * xbel_item_get_type:
 * @item: A item.
 *
 * ...
 *
 * Return value: The type of the given @item.
 **/
XBELitemType xbel_item_get_type(XBELitem* item)
{
    return item->type;
}

/**
 * xbel_item_get_parent:
 * @item: A valid item.
 *
 * ...
 *
 * Return value: The parent folder of the given @item or %NULL.
 **/
XBELitem* xbel_item_get_parent(XBELitem* item)
{
    g_return_val_if_fail(item, NULL);
    return (XBELitem*)item->parent;
}

/**
 * xbel_item_get_title:
 * @item: A valid item.
 *
 * ...
 *
 * Return value: The title of the given @item or %NULL.
 **/
G_CONST_RETURN gchar* xbel_item_get_title(XBELitem* item)
{
    g_return_val_if_fail(!xbel_item_is_separator(item), NULL);
    return item->title;
}

/**
 * xbel_item_get_desc:
 * @item: A valid item.
 *
 * ...
 *
 * Return value: The description of the item or %NULL.
 **/
G_CONST_RETURN gchar* xbel_item_get_desc(XBELitem* item)
{
    g_return_val_if_fail(!xbel_item_is_separator(item), NULL);
    return item->desc;
}

/**
 * xbel_bookmark_get_href:
 * @bookmark: A bookmark.
 *
 * ...
 *
 * Return value: The uri of the @bookmark.
 **/
G_CONST_RETURN gchar* xbel_bookmark_get_href(XBELitem* bookmark)
{
    g_return_val_if_fail(xbel_item_is_bookmark(bookmark), NULL);
    return bookmark->href;
}

/**
 * xbel_item_is_bookmark:
 * @item:   A valid item.
 *
 * ...
 *
 * Return value: %TRUE if the item is a bookmark.
 **/
gboolean xbel_item_is_bookmark(XBELitem* item)
{
    g_return_val_if_fail(item, FALSE);
    return item->type == XBEL_ITEM_BOOKMARK;
}

/**
 * @item:   A valid item.
 * Return value: TRUE if the item is a separator.
 **/
gboolean xbel_item_is_separator(XBELitem* item)
{
    g_return_val_if_fail(item, FALSE);
    return item->type == XBEL_ITEM_SEPARATOR;
}

/**
 * @item:   A valid item.
 * Return value: TRUE if the item is a folder.
 **/
gboolean xbel_item_is_folder(XBELitem* item)
{
    g_return_val_if_fail(item, FALSE);
    return item->type == XBEL_ITEM_FOLDER;
}

/**
 * @folder: A folder.
 * @folded: TRUE if the folder is folded.
 **/
void xbel_folder_set_folded(XBELitem* folder, gboolean folded)
{
    g_return_if_fail(xbel_item_is_folder(folder));
    folder->folded = folded;
}

/**
 * @item: A valid item.
 * @title: A string to use for the title.
 **/
void xbel_item_set_title(XBELitem* item, const gchar* title)
{
    g_return_if_fail(!xbel_item_is_separator(item));
    g_free(item->title);
    item->title = g_strdup(title);
}

/**
 * @item: A valid item.
 * @title: A string to use for the description.
 **/
void xbel_item_set_desc(XBELitem* item, const gchar* desc)
{
    g_return_if_fail(!xbel_item_is_separator(item));
    g_free(item->desc);
    item->desc = g_strdup(desc);
}

/**
 * The uri must not be empty.
 *
 * This uri is not currently validated in any way. However it might be later.
 *
 * @bookmark: A bookmark.
 * @href: A valid uri.
 **/
void xbel_bookmark_set_href(XBELitem* bookmark, const gchar* href)
{
    g_return_if_fail(href);
    g_return_if_fail(xbel_item_is_bookmark(bookmark));
    g_free(bookmark->href);
    bookmark->href = g_strdup(href);
}

gboolean xbel_folder_from_data(XBELitem* folder, const gchar* data
 , GError** error)
{
    g_return_val_if_fail(data, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(xbel_folder_get_n_items(folder))
        g_warning("XBEL: The folder is not empty.");
    xmlDocPtr doc;
    if((doc = xmlParseMemory(data, strlen(data))) == NULL)
    {
        // No valid xml or broken encoding
        *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
        , "Malformed document.");
        return FALSE;
    }
    if(!xbel_folder_from_xmlDocPtr(folder, doc))
    {
        // Parsing failed
        xmlFreeDoc(doc);
        *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
         , "Malformed document.");
        return FALSE;
    }
    xmlFreeDoc(doc);
    return TRUE;
}

gboolean xbel_folder_from_file(XBELitem* folder, const gchar* filename
 , GError** error)
{
    g_return_val_if_fail(filename, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(xbel_folder_get_n_items(folder))
        g_warning("XBEL: The folder is not empty.");
    if(!g_file_test(filename, G_FILE_TEST_EXISTS))
    {
        // File doesn't exist
        *error = g_error_new(G_FILE_ERROR, G_FILE_ERROR_NOENT
         , "File not found.");
        return FALSE;
    }
    xmlDocPtr doc;
    if((doc = xmlParseFile(filename)) == NULL)
    {
        // No valid xml or broken encoding
        *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
         , "Malformed document.");
        return FALSE;
    }
    if(!xbel_folder_from_xmlDocPtr(folder, doc))
    {
        // Parsing failed
        xmlFreeDoc(doc);
        *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
         , "Malformed document.");
        return FALSE;
    }
    xmlFreeDoc(doc);
    return TRUE;
}

gboolean xbel_folder_from_data_dirs(XBELitem* folder, const gchar* filename
 , gchar** path, GError** error)
{
    g_return_val_if_fail(filename, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(xbel_folder_get_n_items(folder))
        g_warning("XBEL: The folder is not empty.");
    // FIXME: Essentially unimplemented

    *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
     , "Malformed document.");
    return FALSE;
}

static gchar* xbel_xml_element(const gchar* name, const gchar* value)
{
    if(!value) return g_strdup("");
    gchar* valueEscaped = g_markup_escape_text(value, -1);
    gchar* XML = g_strdup_printf("<%s>%s</%s>\n", name, valueEscaped, name);
    g_free(valueEscaped);
    return XML;
}

static gchar* xbel_item_to_data(XBELitem* item)
{
    g_return_val_if_fail(item, NULL);
    gchar* XML = NULL;
    switch(xbel_item_get_type(item))
    {
    case XBEL_ITEM_FOLDER:
    {
        GString* _XML = g_string_new(NULL);
        guint n = xbel_folder_get_n_items(item);
        guint i; for(i = 0; i < n; i++)
        {
            XBELitem* _item = xbel_folder_get_nth_item(item, i);
            gchar* itemXML = xbel_item_to_data(_item);
            g_string_append(_XML, itemXML);
            g_free(itemXML);
        }
        gchar* folded = item->folded ? NULL : g_strdup_printf(" folded=\"no\"");
        gchar* title = xbel_xml_element("title", item->title);
        gchar* desc = xbel_xml_element("desc", item->desc);
        XML = g_strdup_printf("<folder%s>\n%s%s%s</folder>\n"
         , folded ? folded : ""
         , title
         , desc
         , g_string_free(_XML, FALSE));
        g_free(folded); g_free(title); g_free(desc);
        break;
        }
    case XBEL_ITEM_BOOKMARK:
    {
        gchar* hrefEscaped = g_markup_escape_text(item->href, -1);
        gchar* href = g_strdup_printf(" href=\"%s\"", hrefEscaped);
        g_free(hrefEscaped);
        gchar* title = xbel_xml_element("title", item->title);
        gchar* desc = xbel_xml_element("desc", item->desc);
        XML = g_strdup_printf("<bookmark%s>\n%s%s%s</bookmark>\n"
         , href
         , title
         , desc
         , "");
        g_free(href); g_free(title); g_free(desc);
        break;
    }
    case XBEL_ITEM_SEPARATOR:
        XML = g_strdup("<separator/>\n");
        break;
    default:
        g_warning("XBEL: Unknown item type");
    }
    return XML;
}

gchar* xbel_folder_to_data(XBELitem* folder, gsize* length, GError** error)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    // FIXME: length is never filled
    GString* innerXML = g_string_new(NULL);
    guint n = xbel_folder_get_n_items(folder);
    guint i; for(i = 0; i < n; i++)
    {
        gchar* sItem = xbel_item_to_data(xbel_folder_get_nth_item(folder, i));
        g_string_append(innerXML, sItem);
        g_free(sItem);
    }
    gchar* title = xbel_xml_element("title", folder->title);
    gchar* desc = xbel_xml_element("desc", folder->desc);
    gchar* outerXML;
    outerXML = g_strdup_printf("%s%s<xbel version=\"1.0\">\n%s%s%s</xbel>\n"
     , "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
     , "<!DOCTYPE xbel PUBLIC \"+//IDN python.org//DTD XML Bookmark Exchange Language 1.0//EN//XML\" \"http://www.python.org/topics/xml/dtds/xbel-1.0.dtd\">\n"
     , title
     , desc
     , g_string_free(innerXML, FALSE));
    g_free(title); g_free(desc);
    return outerXML;
}

gboolean xbel_folder_to_file(XBELitem* folder, const gchar* filename, GError** error)
{
    gchar* data;
    if(!(data = xbel_folder_to_data(folder, NULL, error)))
        return FALSE;
    FILE* fp;
    // FIXME: Should undergo a few tests
    if(!(fp = fopen(filename, "w")))
    {
        *error = g_error_new(G_FILE_ERROR, G_FILE_ERROR_ACCES
         , "Writing failed.");
        return FALSE;
    }
    fputs(data, fp);
    fclose(fp);
    g_free(data);
    return TRUE;
}
